from collections import Counter

import numpy as np
import pandas as pd
from scipy.stats import pearsonr as pearsonr


def extract_regions(X, y, regions=[]):
    """
    Extract data of specific regions
    :param X: pd.DataFrame, input data
    :param y: pd.Series, target data
    :param regions: list, list of regions to extract
    """
    X, y = X.copy(), y.copy()

    indexes = X[X["Region"].isin(regions)].index
    X = X.loc[indexes, :].reset_index(drop=True)
    y = y.loc[indexes].reset_index(drop=True)

    return X, y


def _compute_correlations(X, y):
    feature_correlations = {}

    for feature in X.columns:
        # Ignore columns that are object or categorical type
        if X[feature].dtypes.name in ["object", "category"]:
            continue

        f_series = X[feature]

        # Ignore NaN values in either y and f_series
        mask = ~np.isnan(y) | ~np.isnan(f_series)
        y = y[mask]
        f_series = f_series[mask]

        if np.std(f_series) == 0 or np.std(y) == 0:
            feature_correlations[feature] = np.nan
        else:
            try:
                r = pearsonr(y, f_series)[0]
                feature_correlations[feature] = round(r, 3)
            except:
                feature_correlations[feature] = np.nan

    return feature_correlations


def find_most_common_top_feature(top_feature_by_region):
    """
    Find the most common top feature and number of occurences
    :param top_feature_by_region: dict, top feature by region
    """
    # Extract the first position values from the tuples
    values_first_position = [value[0][0] for value in top_feature_by_region.values()]

    # Count occurrences of each value and average value of the top feature
    counter = Counter(values_first_position)

    # Return the counter which contains the most
    # common top feature(s) and number of occurences
    return counter


def get_top_correlated_features(inputs, targets):
    """
    Get most correlated features for each region
    :param inputs: pd.DataFrame, input data
    :param targets: pd.Series, target data
    :param type: str, Find top feature by correlation (top), or compute all feature correlations (all)
    """
    feature_by_region = {}

    for region_id in inputs["Region"].unique():
        X, y = extract_regions(inputs, targets, regions=[region_id])

        feature_correlations = _compute_correlations(X, y)

        # Exclude any nan values
        feature_correlations = {k: v for k, v in feature_correlations.items() if not np.isnan(v)}

        if not feature_correlations:
            continue

        # Sorts the items of feature_correlations, a dictionary mapping features to their
        # correlation values, by the absolute value of the correlation in descending order,
        # resulting in a list of tuples with the feature and its corresponding correlation value
        sorted_corr_features = sorted(
            feature_correlations.items(), key=lambda x: abs(x[1]), reverse=True
        )

        top_feature, top_corr = sorted_corr_features[0]
        feature_by_region[region_id] = ([top_feature], [top_corr])

    counter = find_most_common_top_feature(feature_by_region)

    return feature_by_region, counter


def get_all_features_correlation(inputs, targets, method):
    """
    Get the top correlated features for each region
    :param inputs: pd.DataFrame, input data
    :param targets: pd.Series, target data
    :param method: str, method to use to find the top correlated features
    """
    frames = []
    for region_id in inputs["Region"].unique():
        X, y = extract_regions(inputs, targets, regions=[region_id])

        feature_correlations = _compute_correlations(X, y)

        # Exclude any nan values
        feature_correlations = {k: v for k, v in feature_correlations.items() if not np.isnan(v)}

        if not feature_correlations:
            continue

        split_keys = []
        for key in feature_correlations.keys():
            parts = key.split(" ")
            cei = parts[0]
            time_period = " ".join(parts[1:])

            split_keys.append([cei, time_period])

        # split_keys = [key.rsplit("_", 1) for key in feature_correlations.keys()]
        values = list(feature_correlations.values())

        # Creating a DataFrame
        df = pd.DataFrame(split_keys, columns=["Metric", method])
        df["Value"] = values

        # If you want to pivot the DataFrame so each metric becomes a column name and include the year as a separate column
        df_pivoted = df.pivot_table(
            index=method, columns="Metric", values="Value", aggfunc="first"
        ).reset_index()
        df_pivoted["Region"] = region_id
        # Move the 'Region' column to the front
        cols = df_pivoted.columns.tolist()
        cols = cols[-1:] + cols[:-1]
        df_pivoted = df_pivoted[cols]

        frames.append(df_pivoted)

    if len(frames):
        feature_by_region = pd.concat(frames)
    else:
        feature_by_region = pd.DataFrame()

    return feature_by_region



import ast
import os
import sqlite3
import warnings
from configparser import ConfigParser
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

import arrow as ar
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import palettable as pal
import pandas as pd
import seaborn as sns
from tqdm import tqdm

from geocif import logger as log
from geocif import utils
from .viz import plot

warnings.simplefilter(action="ignore", category=FutureWarning)


@dataclass
class Geoanalysis:
    path_config_files: List[Path] = field(default_factory=list)
    logger: log = None
    parser: ConfigParser = field(default_factory=ConfigParser)

    def __post_init__(self):
        self.country: str = None
        self.countries: list = None
        self.crop: str = None
        self.table: str = None
        self.forecast_season: int = None
        self.model_names: list = []
        self.df_analysis: pd.DataFrame = None
        self.lag_yield_as_feature: bool = None
        self.number_lag_years: int = None
        self.all_seasons_with_yield: list = None

        self.dir_out = Path(self.parser.get("PATHS", "dir_output"))
        self._date = ar.utcnow().to("America/New_York")
        self.today = self._date.format("MMMM_DD_YYYY")

        self.dir_ml = self.dir_out / "ml"
        self.dir_db = self.dir_ml / "db"
        self.dir_analysis = self.dir_ml / "analysis" / self.today
        os.makedirs(self.dir_db, exist_ok=True)
        os.makedirs(self.dir_analysis, exist_ok=True)

        self.db_forecasts = self.parser.get("DEFAULT", "db")
        self.db_path = self.dir_db / self.db_forecasts

        dir_input = Path(self.parser.get("PATHS", "dir_input"))
        self.dir_shapefiles = dir_input / "Global_Datasets" / "Regions" / "Shps"

    def table_exists(self, db_path, table_name):
        # Create a connection to the SQLite database
        with sqlite3.connect(db_path) as con:
            # Create a cursor object using the cursor() method
            cursor = con.cursor()

            # Define the query to find the table
            query = f"SELECT name FROM sqlite_master WHERE type='table' AND name=?"

            # Execute the prepared query passing the table_name as a parameter
            cursor.execute(query, (table_name,))

            # Fetch the result
            result = cursor.fetchone()

            # Close the cursor
            cursor.close()

        # Return True if a result is found, False otherwise
        return result is not None

    def query(self):
        self.logger.info(f"Query {self.country} {self.crop}")
        con = sqlite3.connect(self.db_path)

        # Read from database, where country and crop match
        query = "SELECT * FROM " + self.table
        try:
            self.df_analysis = pd.read_sql_query(query, con)

            # Select just Country and Crop
            self.df_analysis = self.df_analysis[
                (self.df_analysis["Country"] == self.country)
                & (self.df_analysis["Crop"] == self.crop)
                & (self.df_analysis["Model"] == self.model)
            ]
        except Exception as e:
            pass

        con.commit()
        con.close()

    def annual_metrics(self, df):
        """
        Compute metrics for a given dataframe
        :param df: dataframe containing Observed and Forecast data
        """
        import scipy.stats
        from sklearn.metrics import mean_squared_error, mean_absolute_error

        if len(df) < 3:
            return pd.Series()

        # Compute metrics
        rmse = np.sqrt(mean_squared_error(df[self.observed], df[self.predicted]))
        nse = utils.nse(df[self.observed], df[self.predicted])
        r2 = scipy.stats.pearsonr(df[self.observed], df[self.predicted])[0] ** 2
        mae = mean_absolute_error(df[self.observed], df[self.predicted])
        mape = utils.mape(df[self.observed], df[self.predicted])
        pbias = utils.pbias(df[self.observed], df[self.predicted])

        # Return as a dictionary
        dict_results = {
            "Root Mean Square Error": rmse,
            "Nash-Sutcliff Efficiency": nse,
            "$r^2$": r2,
            "Mean Absolute Error": mae,
            "Mean Absolute\nPercentage Error": mape,
            "Percentage Bias": pbias,
        }

        return pd.Series(dict_results)

    def regional_metrics(self, df):
        # Compute MAPE for each region, compute within this function
        # Compute metrics

        actual, predicted = df[self.observed], df[self.predicted]
        mape = np.mean(np.abs((actual - predicted) / actual)) * 100

        return pd.Series({"Mean Absolute Percentage Error": mape})

    def add_stage_information(self, df):
        """
        Create a new column called Dekad which contains the last dekad
        :param df: dataframe containing the column Stages for which we will compute Dekad information
        """
        for i, row in df.iterrows():
            # Get the latest stage
            stage = row["Stage Name"].split("-")[0]
            df.loc[i, "Date"] = stage

        return df

    def select_top_N_years(self, group, N=5):
        return group.nsmallest(N, "Mean Absolute Percentage Error")

    def analyze(self):
        self.logger.info(f"Analyze {self.country} {self.crop}")
        # Remove rows with missing values in Observed Yield (tn per ha)
        df = self.df_analysis.dropna(subset=["Observed Yield (tn per ha)"])

        if df.empty:
            return pd.DataFrame(), pd.DataFrame()

        # For each Harvest Year, Stages combination, compute
        # RMSE, NSE, R2, MAE, MAPE, PBIAS
        df_metrics = df.groupby(
            ["Country", "Model", "Harvest Year", "Stage Name", "Stage Range"]
        ).apply(self.annual_metrics)

        df_metrics = df_metrics.reset_index()
        # Assign each unique Stage Name a unique integer identifier
        df_metrics["Stage_ID"] = pd.Categorical(df_metrics["Stage Name"]).codes

        # Order by Harvest Year and Number Stages (ascending)
        df_metrics = df_metrics.sort_values(
            by=["Harvest Year", "Stage_ID"], ascending=[True, True]
        )

        # Add columns with the name of the country and crop
        df_metrics["Country"] = self.country
        df_metrics["Crop"] = self.crop

        # Add stage information for plotting
        df_metrics = self.add_stage_information(df_metrics)

        # Rename level_2 to Metric and 0 to Value
        # df_metrics = df_metrics.rename(columns={"level_2": "Metric", 0: "Value"})
        # breakpoint()
        # df_metrics.to_csv(r'D:\Users\ritvik\projects\GEOGLAM\Output\fao\dekad\ml\analysis\February-28-2024\ethiopia_maize\ab1.csv')
        for metric in [
            "Root Mean Square Error",
            # "Nash-Sutcliff Efficiency",
            "$r^2$",
            "Mean Absolute Error",
            "Mean Absolute\nPercentage Error",
            "Percentage Bias",
        ]:
            self.plot_metric(df_metrics, metric)

        cols = [
            "Country",
            "Region",
            "% of total Area (ha)",
            "Model",
            "Crop",
            "Stage Name",
            "Stage Range",
        ]

        # For each Stages combination, compute MAPE
        df_regional_metrics = (
            df.groupby(cols + ["Harvest Year"])
            .apply(self.regional_metrics)
            .reset_index()
        )

        # HACK
        # For each Country, Region, harvest Year combination, select the 10 years with least MAPE
        df_regional_metrics = (
            df_regional_metrics.groupby(["Country", "Region"])
            .apply(lambda x: self.select_top_N_years(x, 10))
            .reset_index(drop=True)
        )

        # Determine average MAPE for each Country, Region, Model, Crop, Stage Name, Stage Range
        df_regional_metrics = (
            df_regional_metrics.groupby(cols)["Mean Absolute Percentage Error"]
            .mean()
            .reset_index()
        )

        # Create an index based on following columns
        cols = [
            "Country",
            "Crop",
            "Model",
            "Harvest Year",
            "Stage Name",
        ]
        df_metrics.index = df_metrics.apply(
            lambda row: "_".join([str(row[col]) for col in cols]), axis=1
        )
        df_metrics.index.set_names(["Index"], inplace=True)

        cols = [
            "Country",
            "Region",
            "Model",
            "Crop",
            "Stage Name",
        ]
        df_regional_metrics.index = df_regional_metrics.apply(
            lambda row: "_".join([str(row[col]) for col in cols]), axis=1
        )
        df_regional_metrics.index.set_names(["Index"], inplace=True)

        # Format with 3 places after the decimal point
        df_metrics = df_metrics.round(3)
        df_regional_metrics = df_regional_metrics.round(3)

        # Store results in database
        con = sqlite3.connect(self.db_path)
        utils.to_db(self.db_path, f"country_metrics", df_metrics)
        utils.to_db(self.db_path, f"regional_metrics", df_regional_metrics)

        con.commit()
        con.close()

    def get_historic_production(self):
        # Read in historic production data
        dir_output = Path(self.parser.get("PATHS", "dir_output"))
        dir_statistics = dir_output / "cei" / "indices" / self.method / "global"
        country = self.country.title().replace("_", " ")
        crop = self.crop.title().replace("_", " ")
        file = dir_statistics / f"{country}_{crop}_statistics_s1_{self.method}.csv"
        df_historic = pd.read_csv(file)

        df_historic = df_historic[["Region", "Harvest Year", "Yield (tn per ha)"]]

        # Drop rows with NaN values
        df_historic = df_historic.dropna()

        # Determine unique years
        years = df_historic["Harvest Year"].unique()

        # Subset dataframe to only include the last years of the dataset
        df_historic = df_historic[df_historic["Harvest Year"].isin(years[-5:])]

        # For each region, compute the % of the total production
        df_historic = (
            df_historic.groupby("Region")["Yield (tn per ha)"]
            .sum()
            .pipe(lambda x: x / x.sum() * 100)
            .to_frame(name="% of total Area (ha)")
            .reset_index()
        )
        # Find median yield for each region
        # df_historic = (
        #     df_historic.groupby("Region")["Yield (tn per ha)"]
        #     .median()
        #     .to_frame(name="Median Yield (tn per ha)")
        #     .reset_index()
        # )

        return df_historic

    def preprocess(self):
        if self.df_analysis.empty:
            return

        # Add a column called N year average that contains the average of the yield of the last 10 years
        # this will be the same for each dekad in any year
        df_lag_yield = self.df_analysis.copy()

        df_lag_yield = (
            df_lag_yield.groupby("Region")["Median Yield (tn per ha)"]
            .median()
            .reset_index()
        )
        df_lag_yield.columns = ["Region", f"{self.number_lag_years} year average"]

        self.df_analysis = self.df_analysis.merge(df_lag_yield, on="Region", how="left")

        df_historic = self.get_historic_production()
        self.df_analysis = self.df_analysis.merge(df_historic, on="Region", how="left")

        # Add a column called anomaly that is the ratio between the predicted yield and the N year average
        self.df_analysis["Anomaly"] = (
            self.df_analysis[self.predicted]
            * 100.0
            / self.df_analysis["Median Yield (tn per ha)"]
        )

        # Compute the yield from the last year
        # Add a column called Ratio Last Year that is the ratio between the predicted yield and the last year yield
        # self.df_analysis["Ratio Last Year"] = (
        #     self.df_analysis[self.predicted]
        #     * 100.0
        #     / self.df_analysis[f"Last Year Yield (tn per ha)"]
        # )

        return self.df_analysis

    def map_regional(self):
        con = sqlite3.connect(self.db_path)

        # Read from database, where country and crop match
        query = "SELECT * FROM country_metrics"
        df_country = pd.read_sql_query(query, con)
        query = "SELECT * FROM regional_metrics"
        df_regional = pd.read_sql_query(query, con)

        # Plot a histogram of the MAPE, different color for each country
        # Plotting the histograms with KDE for each country
        df_regional["Country"] = (
            df_regional["Country"].str.replace("_", " ").str.title()
        )
        df_regional["Model"] = df_regional["Model"].str.title()

        # Plotting the histogram with a smaller bin size for greater detail
        # Plotting the KDE for each country, ensuring each step works
        models = df_regional["Model"].unique()
        for model in models:
            df_model = df_regional[df_regional["Model"] == model]

            # HACK: Drop rows where '% of total Area (ha)' is less than 1% and Mean Absolute Percentage Error is > 50%
            # or where the Mean Absolute Percentage Error is greater than 50% if the '% of total Area (ha)' is greater than 1%
            df_tmp = df_model[
                (df_model["% of total Area (ha)"] < 1)
                & (df_model["Mean Absolute Percentage Error"] > 50)
                & (df_model["Country"].isin(["Angola", "United Republic Of Tanzania"]))
            ]

            # Remove df_tmp from df_model
            df_model = df_model.drop(df_tmp.index)
            # Plot the histogram of MAPE
            # Create bins for '% of total Area (ha)' and 'MAPE'
            bin_edges = np.linspace(0, df_model["% of total Area (ha)"].max() + 1, 5 + 1)

            df_model["Area Bins"] = pd.cut(
                df_model["% of total Area (ha)"],
                bins=bin_edges,
                precision=0,
            )
            df_model["MAPE Bins"] = pd.cut(
                df_model["Mean Absolute Percentage Error"],
                bins=5,  # [0, 5, 10, 15, 20, 25, 30, 50, max(df_model["Mean Absolute Percentage Error"])],
                right=False,
                precision=1,
            )

            # Count occurrences of MAPE values for each area bin
            area_mape_counts = (
                df_model.groupby(["Area Bins", "MAPE Bins"])
                .size()
                .unstack(fill_value=0)
            )

            # Create the heatmap
            plt.figure(figsize=(10, 8))
            ax = sns.heatmap(
                area_mape_counts,
                annot=True,
                square=True,
                cmap=pal.scientific.sequential.Bamako_20_r.mpl_colormap,
                fmt="d",
            )
            # Do not color or annotate cells with 0
            for text in ax.texts:
                if text.get_text() == "0":
                    text.set_text("")
                    text.set_color("white")

            # plt.title("Heatmap of MAPE Bins vs % Total Area Bins")
            plt.ylabel("% of Total Area (ha) Bins")
            plt.xlabel("MAPE Bins")

            # Adjust y-axis labels to horizontal
            ax.set_yticklabels(ax.get_yticklabels(), rotation=0)

            # Invert y-axis to have the highest bin at the top
            ax.invert_yaxis()
            plt.savefig(self.dir_analysis / f"heatmap_{model}.png", dpi=250)
            plt.close()

            # Plot the KDE of MAPE
            plt.figure(figsize=(12, 8))
            for label, group_data in df_model.groupby("Country"):
                sns.kdeplot(
                    group_data["Mean Absolute Percentage Error"],
                    label=label,
                    clip=(0, None),
                    # bins=len(group_data),
                    # kde=True,
                )

            # Add minor ticks on the x-axis
            plt.minorticks_on()

            # Setting the title and labels
            plt.title(
                f"Kernel Density Estimation of Mean Absolute Percentage Error by Country - {model}"
            )
            plt.xlabel("Mean Absolute Percentage Error (%)")
            plt.ylabel("Density")
            plt.legend(title="Country", title_fontsize="13")
            plt.savefig(self.dir_analysis / f"mape_histogram_{model}.png", dpi=250)
            plt.close()

            # Map MAPE at regional level
            df_model["Country Region"] = (
                df_model["Country"].str.lower().str.replace("_", " ")
                + " "
                + df_model["Region"].str.lower()
            )

            fname = f"mape_{self.crop}_{model}.png"
            col = "Mean Absolute Percentage Error"
            countries = df_model["Country"].unique().tolist()
            # make it title case and replace _ with space
            countries = [country.title().replace("_", " ") for country in countries]
            countries = ["Malawi"]
            df_model = df_model[df_model["Country"].isin(countries)]
            self.dg = self.dg[self.dg["ADM0_NAME"].isin(countries)]
            plot.plot_df_shpfile(
                self.dg,  # dataframe containing adm1 name and polygon
                df_model,  # dataframe containing information that will be mapped
                merge_col="Country Region",  # Column on which to merge
                name_country=countries,
                name_col=col,  # Which column to plot
                dir_out=self.dir_analysis,  # Output directory
                fname=fname,  # Output file name
                label=f"MAPE (%)",
                vmin=df_model[col].min(),
                vmax=df_model[col].max(),
                cmap=pal.scientific.sequential.Bamako_20_r,
                series="sequential",
                show_bg=False,
                annotate_regions=False,
                loc_legend="lower left",
            )

        con.commit()
        con.close()

    def map(self, df_plot):
        # df_plot = self.df_analysis.copy()
        models = df_plot["Model"].unique()

        for model in models:
            df_model = df_plot[df_plot["Model"] == model]

            countries = ["malawi"]
            df_model = df_model[df_model["Country"].isin(countries)]
            self.dg = self.dg[self.dg["ADM0_NAME"].isin(["Malawi", "malawi"])]

            countries = df_model["Country"].unique().tolist()
            if len(countries) > 1:
                self.dir_plot = self.dir_analysis
                fname_prefix = f"{len(countries)}_countries"
            else:
                self.dir_plot = self.dir_analysis / self.country / self.crop
                fname_prefix = f"{self.country}"
            countries = [country.title().replace("_", " ") for country in countries]
            df_model["Country Region"] = (
                df_model["Country"].str.lower().str.replace("_", " ")
                + " "
                + df_model["Region"].str.lower().str.replace("_", " ")
            )


            # Change Harvest year to type int
            df_model["Harvest Year"] = df_model["Harvest Year"].astype(int)
            annotate_region_column = (
                "ADM1_NAME" if self.admin_zone == "admin_1" else "ADM2_NAME"
            )
            analysis_years = df_model["Harvest Year"].unique()
            for idx, year in enumerate(tqdm(analysis_years, desc="Map")):
                df_harvest_year = df_model[df_model["Harvest Year"] == year]

                for time_period in tqdm(
                    df_harvest_year["Stage Name"].unique(), desc="Map"
                ):
                    df_time_period = df_harvest_year[
                        df_harvest_year["Stage Name"] == time_period
                    ]
    #
    #                 """ % of total area """
                    if idx == 0:
                        fname = f"{self.country}_{self.crop}_perc_area.png"
                        col = "% of total Area (ha)"
                        plot.plot_df_shpfile(
                            self.dg,  # dataframe containing adm1 name and polygon
                            df_model,  # dataframe containing information that will be mapped
                            merge_col="Country Region",  # Column on which to merge
                            name_country=countries,  # Plot global map
                            name_col=col,  # Which column to plot
                            dir_out=self.dir_plot / str(year),  # Output directory
                            fname=fname,  # Output file name
                            label=f"% of Total Area (ha)\n{self.crop.title()}",
                            vmin=df_model[col].min(),
                            vmax=df_model[col].max(),
                            cmap=pal.scientific.sequential.Bamako_20_r,
                            series="sequential",
                            show_bg=False,
                            annotate_regions=False,
                            annotate_region_column=annotate_region_column,
                            loc_legend="lower left",
                        )
    #
    #                 #     """ Unique regions """
                    fname = f"{self.country}_{self.crop}_region_ID.png"
                    col = "Region_ID"
                    df_model[col] = df_model[col].astype(int) + 1
                    if len(df_model["Region_ID"].unique() > 1):
                        # Create a dictionary with each region assigned a unique integer identifier and name
                        dict_region = {
                            int(key): key for key in df_time_period["Region_ID"].unique()
                        }
                        plot.plot_df_shpfile(
                            self.dg,  # dataframe containing adm1 name and polygon
                            df_model,  # dataframe containing information that will be mapped
                            dict_lup=dict_region,
                            merge_col="Country Region",  # Column on which to merge
                            name_country=countries,  # Plot global map
                            name_col=col,  # Which column to plot
                            dir_out=self.dir_plot / str(year),  # Output directory
                            fname=fname,  # Output file name
                            label=f"Region Cluster\n{self.crop.title()}",
                            vmin=df_model[col].min(),
                            vmax=df_model[col].max(),
                            cmap=pal.tableau.Tableau_20.mpl_colors,
                            series="qualitative",
                            show_bg=False,
                            alpha_feature=1,
                            use_key=True,
                            annotate_regions=False,
                            annotate_region_column=annotate_region_column,
                            loc_legend="lower left",
                        )
    #                     breakpoint()

                    # """ Anomaly """
                    # fname = (
                    #     f"{fname_prefix}_{self.crop}_{time_period}_{year}_anomaly.png"
                    # )
                    # plot.plot_df_shpfile(
                    #     self.dg,  # dataframe containing adm1 name and polygon
                    #     df_harvest_year,  # dataframe containing information that will be mapped
                    #     merge_col="Country Region",  # Column on which to merge
                    #     name_country=countries,  # Plot global map
                    #     name_col="Anomaly",  # Which column to plot
                    #     dir_out=self.dir_plot / str(year),  # Output directory
                    #     fname=fname,  # Output file name
                    #     label=f"% of {self.number_lag_years}-year Median Yield\n{self.crop.title()}, {year}",
                    #     vmin=df_harvest_year["Anomaly"].min(),
                    #     vmax=110,  # df_harvest_year["Anomaly"].max(),
                    #     cmap=pal.cartocolors.diverging.Geyser_5_r,
                    #     series="sequential",
                    #     show_bg=False,
                    #     annotate_regions=False,
                    #     annotate_region_column=annotate_region_column,
                    #     loc_legend="lower left",
                    # )

                    """ Predicted Yield """
                    fname = f"{fname_prefix}_{self.crop}_{time_period}_{year}_predicted_yield.png"
                    plot.plot_df_shpfile(
                        self.dg,  # dataframe containing adm1 name and polygon
                        df_harvest_year,  # dataframe containing information that will be mapped
                        merge_col="Country Region",  # Column on which to merge
                        name_country=countries,  # Plot global map
                        name_col="Predicted Yield (tn per ha)",  # Which column to plot
                        dir_out=self.dir_plot / str(year),  # Output directory
                        fname=fname,  # Output file name
                        label=f"Predicted Yield (Mg/ha)\n{self.crop.title()}, {year}",
                        vmin=df_harvest_year[self.predicted].min(),
                        vmax=df_harvest_year[self.predicted].max(),
                        cmap=pal.scientific.sequential.Bamako_20_r,
                        series="sequential",
                        show_bg=False,
                        annotate_regions=False,
                        annotate_region_column=annotate_region_column,
                        loc_legend="lower left",
                    )

                    """ Ratio of Predicted to last Year Yield """
                    # fname = f"{self.country}_{self.crop}_{time_period}_{year}_ratio_last_year_yield.png"
                    # plot.plot_df_shpfile(
                    #     self.dg,  # dataframe containing adm1 name and polygon
                    #     df_time_period,  # dataframe containing information that will be mapped
                    #     merge_col="Country Region",  # Column on which to merge
                    #     name_country=countries,  # Plot global map
                    #     name_col="Ratio Last Year",  # Which column to plot
                    #     dir_out=self.plot_dir / str(year),  # Output directory
                    #     fname=fname,  # Output file name
                    #     label=f"Ratio Last Year to {self.predicted}\n{self.crop.title()}, {time_period} {year}",
                    #     vmin=df_time_period["Ratio Last Year"].min(),
                    #     vmax=df_time_period["Ratio Last Year"].max(),
                    #     cmap=pal.scientific.sequential.Bamako_20_r,
                    #     series="sequential",
                    #     show_bg=False,
                    #     annotate_regions=False,
                    #     annotate_region_column=annotate_region_column,
                    #     loc_legend="lower left",
                    # )

                    # Area
                    # breakpoint()
                    # if df_time_period["Area (ha)"].notna().all():
                    #     fname = f"{self.country}_{self.crop}_{year}_area.png"
                    #     plot.plot_df_shpfile(
                    #         self.dg,  # dataframe containing adm1 name and polygon
                    #         df_time_period,  # dataframe containing information that will be mapped
                    #         merge_col="Country Region",  # Column on which to merge
                    #         name_country=country,  # Plot global map
                    #         name_col="Area (ha)",  # Which column to plot
                    #         dir_out=self.plot_dir / str(year),  # Output directory
                    #         fname=fname,  # Output file name
                    #         label=f"{self.predicted}\n{self.crop.title()}, {time_period}",
                    #         vmin=df_time_period[self.predicted].min(),
                    #         vmax=df_time_period[self.predicted].max(),
                    #         cmap=pal.scientific.sequential.Bamako_20_r,
                    #         series="sequential",
                    #         show_bg=False,
                    #         annotate_regions=True,
                    #         loc_legend="lower left",
                    #     )

    def plot_metric(self, df, metric="$r^2$"):
        with plt.style.context("science"):
            fig, ax = plt.subplots(figsize=(10, 5))
            ax = sns.lineplot(data=df, x="Date", y=metric, ax=ax)  # "$r^2$"
            ax.set_xlabel("")
            ax.set_ylabel(metric)
            plt.xticks(rotation=0)
            plt.tight_layout()

            # If metric is $r^2$ or NSE, do not plot values below 0
            if metric in ["$r^2$", "Nash-Sutclwiiff Efficiency"]:
                plt.ylim(0, 1)

            # Replace \n in metric
            metric = metric.replace("\n", " ")
            fname = f"{self.country}_{self.crop}_{metric}.png"

            plt.savefig(self.dir_analysis / fname, dpi=250)
            plt.close()

    def execute(self):
        self.query()
        aa = self.preprocess()
        self.analyze()

        return aa

    def get_config_data(self):
        try:
            with sqlite3.connect(self.db_path) as con:
                # Find names of all tables starting with 'config'
                query = "SELECT name FROM sqlite_master WHERE type='table' AND name LIKE 'config%'"
                df = pd.read_sql_query(query, con)

                if df.empty:
                    raise ValueError("No configuration tables found")

                # Extract datetime from the table names
                re = "(\d{4} \d{2}:\d{2})$"
                df["datetime"] = pd.to_datetime(
                    df["name"].str.extract(re)[0], format="%Y %H:%M"
                )

                # Sort the DataFrame by datetime in descending order and get the latest config file
                latest_config = df.sort_values(by="datetime", ascending=False).iloc[0][
                    "name"
                ]

                self.logger.info("=====================================")
                self.logger.info(f"\t{latest_config}")
                self.logger.info("=====================================")
                # Read the latest config file
                query = f"SELECT * FROM {latest_config}"
                self.df_config = pd.read_sql_query(query, con)
        except Exception as e:
            self.logger.error(f"Failed to get configuration data: {e}")

    def setup(self):
        """

        Args:
            country:
            crop:
            model:

        Returns:

        """
        self.dict_config = {}
        self.get_config_data()

        self.observed = "Observed Yield (tn per ha)"
        self.predicted = "Predicted Yield (tn per ha)"

        # Get the ML section
        df_ml = self.df_config[self.df_config["Section"] == "ML"]

        self.countries = ast.literal_eval(
            df_ml[df_ml["Option"] == "countries"]["Value"].values[0]
        )
        for country in self.countries:
            df = self.df_config[self.df_config["Section"] == country]

            method = df[df["Option"] == "method"]["Value"].values[0]
            crops = ast.literal_eval(df[df["Option"] == "crops"]["Value"].values[0])
            models = ast.literal_eval(df[df["Option"] == "models"]["Value"].values[0])
            admin_zone = df[df["Option"] == "admin_zone"]["Value"].values[0]
            name_shapefile = df[df["Option"] == "boundary_file"]["Value"].values[0]

            for crop in crops:
                # Does a table with the name {country}_{crop} exist in the database?
                table = f"{country}_{crop}"
                if self.table_exists(self.db_path, table):
                    self.dict_config[f"{country}_{crop}"] = {
                        "method": method,
                        "crops": crop,
                        "models": models,
                        "admin_zone": admin_zone,
                        "name_shapefile": name_shapefile,
                    }

        self.dg = gpd.read_file(
            self.dir_shapefiles / "adm_shapefile.shp",
            engine="pyogrio",
        )

        # Hack rename Tanzania to United Republic of Tanzania
        self.dg["ADMIN0"] = self.dg["ADMIN0"].replace(
            "Tanzania", "United Republic of Tanzania"
        )

        # Rename ADMIN0 to ADM0_NAME and ADMIN1 to ADM1_NAME and ADMIN2 to ADM2_NAME
        self.dg = self.dg.rename(
            columns={
                "ADMIN0": "ADM0_NAME",
                "ADMIN1": "ADM1_NAME",
                "ADMIN2": "ADM2_NAME",
            }
        )

        # Create a new column called Country Region that is the concatenation of ADM0_NAME and ADM1_NAME
        # however if ADM2_NAME is not null, then it is the concatenation of ADM0_NAME and ADM2_NAME
        self.dg["Country Region"] = self.dg["ADM0_NAME"]
        self.dg["Country Region"] = self.dg["Country Region"].str.cat(
            self.dg["ADM1_NAME"], sep=" "
        )
        self.dg.loc[self.dg["ADM2_NAME"].notna(), "Country Region"] = (
            self.dg["ADM0_NAME"] + " " + self.dg["ADM2_NAME"]
        )
        # Make it lower case
        self.dg["Country Region"] = self.dg["Country Region"].str.lower().replace("_", " ")


def run(path_config_files=[Path("../config/geocif.txt")]):
    logger, parser = log.setup_logger_parser(path_config_files)
    obj = Geoanalysis(path_config_files, logger, parser)
    obj.setup()

    """ Loop over each country, crop, model combination in dict_config """
    frames = []
    for country_crop, value in obj.dict_config.items():
        obj.crop = value["crops"]
        # to get country, remove obj.crops from country_crop
        obj.country = country_crop.replace(f"_{obj.crop}", "")

        obj.admin_zone = value["admin_zone"]
        obj.boundary_file = value["name_shapefile"]
        obj.method = value["method"]
        obj.number_lag_years = 5

        obj.table = f"{obj.country}_{obj.crop}"
        models = value["models"]
        for model in models:
            obj.model = model

            df_tmp = obj.execute()
            frames.append(df_tmp)

    dk = pd.concat(frames)

    # Map the metrics
    # obj.map_regional()
    obj.map(dk)


if __name__ == "__main__":
    run()



import ast
import multiprocessing as mp
import os
from configparser import ConfigParser
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

import arrow as ar
import geopandas as gp
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn
from tqdm import tqdm

from geocif import logger as log
from .cei import definitions as di
from .ml import correlations
from .ml import feature_engineering as fe
from .ml import feature_selection as fs
from .ml import output
from .ml import stages
from .ml import stats
from .ml import trainers
from .ml import trend
from .ml import xai

plt.style.use("default")
sklearn.set_config(transform_output="pandas")

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)


class TimeFilter:
    def __init__(self, threshold=1):  # Time threshold in seconds
        self.threshold = threshold

    def __call__(self, frame):
        return frame.time() > self.threshold


class TQDMCallback:
    def __init__(self, total_iterations):
        self.pbar = tqdm(total=total_iterations)

    def after_iteration(self, info):
        self.pbar.update(1)


@dataclass
class Geocif:
    method: str = "dekad_r"
    group_by: List[str] = field(
        default_factory=lambda: ["Index", "Country", "Region", "Crop", "Season"]
    )
    metrics: List[str] = field(default_factory=lambda: ["$r^2$", "RMSE", "MAE", "MAPE"])
    logger: log = None
    parser: ConfigParser = field(default_factory=ConfigParser)

    def __post_init__(self):
        self.dir_output = Path(self.parser.get("PATHS", "dir_output"))
        self.dir_condition = Path(self.parser.get("PATHS", "dir_condition"))
        os.makedirs(self.dir_output / "ml", exist_ok=True)

        self.country: str = None
        self.crop: str = None
        self.forecast_season: int = None
        self.all_stages: list = []
        self.all_seasons: list = []
        self.all_seasons_with_yield: list = []
        self.model_names: list = []
        self.feature_names: list = []
        self.selected_features: list = []
        self.num_cpus = mp.cpu_count()

        self._date = ar.utcnow().to("America/New_York")
        self.today = self._date.format("MMMM_DD_YYYY")
        self.today_year = self._date.year
        self.today_doy = int(self._date.format("DDD"))
        self.today_full = self._date.format("MMMM_DD_YYYY_HH_mm")

        self.df_forecast = pd.DataFrame()
        """
        ====================================================================
                                Config file: Default
        ====================================================================
        """
        self.method = self.parser.get("DEFAULT", "method")
        self.db_forecasts = self.parser.get("DEFAULT", "db")
        self.countries = ast.literal_eval(self.parser.get("DEFAULT", "countries"))
        self.do_parallel = self.parser.getboolean("DEFAULT", "do_parallel")
        self.update_input_file = self.parser.getboolean("DEFAULT", "update_input_file")
        self.correlation_plots = self.parser.getboolean("DEFAULT", "correlation_plots")
        self.national_correlation = self.parser.getboolean(
            "DEFAULT", "national_correlation"
        )
        self.correlation_plot_groupby = self.parser.get(
            "DEFAULT", "correlation_plot_groupby"
        )
        self.run_ml = self.parser.getboolean("DEFAULT", "run_ml")

        """
        ====================================================================
                                Config file: ML
        ====================================================================
        """
        self.model_type = self.parser.get("ML", "model_type")
        self.fraction_simulate = self.parser.getint("ML", "fraction_simulate")
        self.analogous_year_yield_as_feature = self.parser.getboolean(
            "ML", "analogous_year_yield_as_feature"
        )
        self.last_year_yield_as_feature = self.parser.getboolean(
            "ML", "last_year_yield_as_feature"
        )
        self.panel_model = self.parser.getboolean("ML", "panel_model")
        self.panel_model_region = self.parser.get("ML", "panel_model_region")
        self.use_outlook_as_feature = self.parser.getboolean(
            "ML", "use_outlook_as_feature"
        )
        self.lag_yield_as_feature = self.parser.getboolean("ML", "lag_yield_as_feature")
        self.number_median_years = self.parser.getint("ML", "median_years")
        self.median_yield_as_feature = self.parser.getboolean(
            "ML", "median_yield_as_feature"
        )
        self.number_lag_years = self.parser.getint("ML", "lag_years")
        self.cluster_strategy = self.parser.get("ML", "cluster_strategy")
        self.feature_selection = self.parser.get("ML", "feature_selection")
        self.check_yield_trend = self.parser.getboolean("ML", "check_yield_trend")
        self.run_latest_time_period = self.parser.getboolean(
            "ML", "run_latest_time_period"
        )
        self.run_every_time_period = self.parser.get("ML", "run_every_time_period")
        self.cat_features: list = ast.literal_eval(
            self.parser.get("ML", "cat_features")
        )

        """
        ====================================================================
                                Variables, Paths
        ====================================================================
        """
        # If ML model is run for individual region or cluster, then Region_ID is the same for each region
        # or cluster and therefore redundant for the ML model
        if self.cluster_strategy in ["individual", "auto_detect"]:
            self.cat_features.remove("Region_ID")

        self.fixed_columns: list = [
            "Country",
            "Region",
            "Crop",
            "Area",
            "Season",
            "Harvest Year",
        ]
        self.target: str = "Yield (tn per ha)"
        self.statistics_columns: list = [
            "Area (ha)",
            "Production (tn)",
        ]

        self.combined_dict = {
            **di.dict_indices,
            **di.dict_ndvi,
            **di.dict_gcvi,
            **di.dict_esi4wk,
            **di.dict_hindex,
        }

        self.combined_keys = list(self.combined_dict.keys())

        self.dir_ml = self.dir_output / "ml"
        self.dir_db = self.dir_ml / "db"
        self.dir_analysis = self.dir_ml / "analysis" / self.today
        dir_input = Path(self.parser.get("PATHS", "dir_input"))
        self.dir_shapefiles = dir_input / "Global_Datasets" / "Regions" / "Shps"

        os.makedirs(self.dir_db, exist_ok=True)
        os.makedirs(self.dir_analysis, exist_ok=True)

        self.db_path = self.dir_db / self.db_forecasts
        output.config_to_db(self.db_path, self.parser, self.today_full)

        # self.pickle_file = self.base_dir / self.parser.get("outlook", "pickle_file")
        # obj_pickle = outlook.Outlook(self.pickle_file)
        # self.df_outlook = obj_pickle.read_outlook_file()

    def train(self, df_region, scaler=None):
        """

        Args:
            df_region:
            scaler:

        Returns:

        """

        """ Perform feature selection """
        target_col = (
            f"Detrended {self.target}" if self.check_yield_trend else self.target
        )

        X_train = df_region[self.feature_names]
        # Drop any columns with NaNs
        X_train = X_train.dropna(axis=1, how="any")
        y_train = df_region[target_col]

        if self.ml_model:
            self.logger.info(f"Selecting features for {self.country} {self.crop}")
            selector, _, self.selected_features = fs.select_features(
                X_train, y_train, method=self.feature_selection
            )
            self.logger.info(f"Selected features: {self.selected_features}")

            """ Update model to include conformal estimates """
            X_train = df_region[self.selected_features + self.cat_features]
            dir_output = (
                self.dir_analysis / self.country / self.crop / self.model_name / str(self.forecast_season)
            )

            region_id = df_region["Region_ID"].unique()[0]
            X_train.to_csv(dir_output / f"X_train_{region_id}.csv", index=False)
            if scaler:
                X_train_nocat = X_train.drop(
                    columns=[
                        item for item in self.cat_features if item != "Harvest Year"
                    ]
                )
                X_train_scaled = scaler.fit_transform(X_train_nocat)
            else:
                X_train_scaled = X_train

            """ Train model """
            self.best_hyperparams, self.model = trainers.auto_train(
                self.cluster_strategy,
                self.model_name,
                False,
                "Harvest Year",
                df_region[self.selected_features + self.cat_features + [self.target]],
                X_train_scaled,
                y_train,
                feature_names=self.selected_features,
                target_col=self.target,
                optimize=self.optimize,
                fraction_loocv=self.fraction_loocv,
                cat_features=self.cat_features,
                monotonic_fatures=[],
                seed=0,
            )

            """ Estimate CI only if flag is True """
            if self.estimate_ci:
                if self.estimate_ci_for_all or self.forecast_season == self.today_year:
                    self.model = trainers.estimate_ci(self.model)

            try:
                if self.model_name == "catboost":
                    self.model.fit(
                        X_train,
                        y_train,
                        cat_features=self.cat_features,
                        verbose=False,
                        # callbacks=[TQDMCallback(self.best_hyperparams["iterations"])],
                    )
                elif self.model_name == "merf":
                    Z_train = np.ones((len(X_train), 1))
                    clusters_train = df_region["Region"]
                    clusters_train.reset_index(drop=True, inplace=True)

                    self.model.fit(X_train, Z_train, clusters_train.astype("object"), y_train.values)
                elif self.model_name == "linear":
                    self.model.fit(X_train_scaled, y_train)
                elif self.model_name == "gam":
                    self.model.fit(X_train_scaled.values, y_train.values)
                    self.best_hyperparams = {}
                elif self.model_name in ["cubist"]:
                    self.model.fit(X_train, y_train)
            except:
                self.logger.error(f"Error fitting model for {self.country} {self.crop}")

    def predict(self, df_region, scaler=None):
        """
        Predict yield for the current stage
        :param df_region:
        :param scaler:
        """
        """ Select dataframe for prediction"""
        X_test = df_region[self.selected_features + self.cat_features]
        y_test = df_region[self.target].values

        """ Do prediction based on model type"""
        if not self.ml_model:
            best_hyperparameters = np.nan
            if self.model_name == "analog":
                y_pred = np.full(len(X_test), df_region["Analogous Year Yield"].values)
            elif self.model_name == "median":
                y_pred = np.full(len(X_test), df_region[f"Median {self.target}"].values)
            elif self.model_name == "last_year":
                y_pred = np.full(
                    len(X_test), df_region[f"Last Year {self.target}"].values
                )
        else:
            if self.model_name in ["linear", "gam"]:
                # Drop cat_features from X_test
                X_test = X_test.drop(
                    columns=[
                        item for item in self.cat_features if item != "Harvest Year"
                    ]
                )
                X_test = scaler.transform(X_test)
                best_hyperparameters = {}

            if self.estimate_ci:
                if self.estimate_ci_for_all or self.forecast_season == self.today_year:
                    y_pred, y_pred_ci = self.model.predict(X_test, alpha=0.1)
                    best_hyperparameters = self.model.get_params().copy()
            elif self.model_name == "merf":
                Z_test = np.ones((len(X_test), 1))
                clusters_test = df_region["Region"]
                clusters_test.reset_index(drop=True, inplace=True)

                y_pred = self.model.predict(X_test, Z_test, clusters_test.astype("object"))
                best_hyperparameters = self.model.fe_model.get_params().copy()
            else:
                y_pred = self.model.predict(X_test)
                best_hyperparameters = self.model.get_params().copy()

        if self.check_yield_trend:
            # Get information for retrending
            for idx, region in enumerate(df_region["Region"].unique()):
                mask_region = self.df_train["Region"] == df_region["Region"].unique()[0]
                df_tmp = self.df_train[mask_region]

                obj_trend = trend.DetrendedData(
                    df_tmp[f"Detrended {self.target}"],
                    df_tmp["Detrend Model"],
                    df_tmp["Detrend Model Type"],
                )

                # Retrend the predicted yield
                y_pred[idx] += trend.compute_trend(
                    obj_trend, df_region.iloc[idx][["Harvest Year"]]
                )[0]

        # Create a dataframe with forecast results
        shp = len(X_test)
        experiment_id = f"{self.country}_{self.crop}"
        now = ar.utcnow().to("America/New_York").format("MMMM-DD-YYYY HH:mm:ss")
        selected_features = self.selected_features + self.cat_features
        df = pd.DataFrame(
            {
                "Experiment_ID": np.full(shp, experiment_id),
                "Date": np.full(shp, self.today),
                "Time": np.full(shp, now),
                "Country": np.full(shp, self.country),
                "Crop": np.full(shp, self.crop),
                "Cluster Strategy": np.full(shp, self.cluster_strategy),
                "Frequency": np.full(shp, self.method),
                "Selected Features": [selected_features.copy() for _ in range(shp)],
                "Best Hyperparameters": np.full(shp, best_hyperparameters),
                "Stage_ID": np.full(shp, self.stage_info["Stage_ID"]),
                "Stage Range": np.full(shp, self.stage_info["Stage Range"]),
                "Stage Name": np.full(shp, self.stage_info["Stage Name"]),
                "Starting Stage": np.full(shp, self.stage_info["Starting Stage"]),
                "Ending Stage": np.full(shp, self.stage_info["Ending Stage"]),
                "Model": np.full(shp, self.model_name),
                "Area (ha)": df_region["Area (ha)"].values,
                "Region_ID": df_region["Region_ID"].values,
                "Region": df_region["Region"].values,
                "Harvest Year": df_region["Harvest Year"].values,
                f"Observed {self.target}": np.around(y_test, 3).ravel(),
                f"Predicted {self.target}": np.around(y_pred, 3).ravel(),
            }
        )

        if self.median_yield_as_feature:
            # Add median yield to dataframe
            df.loc[:, f"Median {self.target}"] = np.around(
                df_region[f"Median {self.target}"].values, 3
            )

        if self.estimate_ci:
            if self.estimate_ci_for_all or self.forecast_season == self.today_year:
                # Iterate over each element in y_pred_ci
                for idx, ci in enumerate(y_pred_ci):
                    # Flatten the list
                    y_pred_ci_ = [item for sublist in ci for item in sublist]
                    # Add to dataframe
                    df.loc[idx, f"lower CI"] = np.around(y_pred_ci_[0], 3)
                    df.loc[idx, f"upper CI"] = np.around(y_pred_ci_[1], 3)

        if self.check_yield_trend:
            df.loc[:, "Detrended Model Type"] = df_region["Detrended Model Type"].values
            df.loc[:, "Detrended Model"] = df_region["Detrended Model"].values

        if self.last_year_yield_as_feature:
            # Add last year yield to dataframe
            df.loc[:, f"Last Year {self.target}"] = np.around(
                df_region[f"Last Year {self.target}"].values, 3
            )

        if self.analogous_year_yield_as_feature:
            try:
                # Add analogous year and yield to dataframe
                df.loc[:, "Analogous Year"] = df_region["Analogous Year"].values
                df.loc[:, "Analogous Year Yield"] = np.around(
                    df_region["Analogous Year Yield"].values, 3
                )
            except:
                breakpoint()

        for col in [
            f"Median {self.target}",
            "Analogous Year",
            "Analogous Year Yield",
            "Detrended Model Type",
            "Detrended Model",
        ]:
            if col not in df.columns:
                df.loc[:, col] = np.NaN

        # Create an index based on following columns
        index_columns = [
            "Model",
            "Country",
            "Region",
            "Crop",
            "Harvest Year",
            "Stage Name",
        ]
        df.index = df.apply(
            lambda row: "_".join([str(row[col]) for col in index_columns]), axis=1
        )

        # name the index level
        df.index.set_names(["Index"], inplace=True)

        return experiment_id, df

    def create_feature_names(self, stages_features, selected_features):
        """

        Args:
            stages_features:
            selected_features:

        Returns:

        """
        # Clear out feature names
        self.feature_names = []

        """ Select stages that will be used for ML
         1. method = "latest" - Select the latest stage
         2. method = "fraction" - Select a fraction (1-100) of all stages
        """
        stages_features = stages.select_stages_for_ml(
            stages_features, method="fraction", n=60
        )

        for stage in stages_features:
            # Convert each element of stage to str and join with _
            _stage = "_".join([str(x) for x in stage])

            # Create a list appending _stage to each element of combined_keys
            _tmp = [f"{col}_{_stage}" for col in self.combined_keys]

            for _t in _tmp:
                parts = _t.split("_")
                cei = parts[0] if parts[1].isdigit() else "_".join(parts[:2])

                # Check if any element of dict_selected_features is in _t
                for x in selected_features:
                    if x not in cei:
                        continue

                    dict_fn = stages.get_stage_information_dict(_t, self.method)
                    tmp_col = dict_fn["CEI"] + " " + dict_fn["Stage Name"]

                    if tmp_col in self.df_train.columns:
                        self.feature_names.append(tmp_col)

        self.feature_names = list(set(self.feature_names))

        if self.median_yield_as_feature:
            self.feature_names.append(f"Median {self.target}")

        if self.lag_yield_as_feature:
            # For the number of years specified in self.number_lag_years
            for i in range(1, self.number_lag_years + 1):
                self.feature_names.append(f"t -{i} {self.target}")

        if self.analogous_year_yield_as_feature:
            self.feature_names.append("Analogous Year")
            self.feature_names.append("Analogous Year Yield")

        if self.use_outlook_as_feature:
            self.feature_names.append("FCST")

        self.selected_features = []

    def loop_ml(self, stages, dict_selected_features, dict_best_cei):
        """

        Args:
            stages:
            dict_selected_features:
            dict_best_cei:

        Returns:

        """
        from sklearn.preprocessing import StandardScaler

        scaler = StandardScaler() if self.model_name in ["linear", "gam"] else None

        """ Train, Predict, Explain and Store results for each region """
        pbar = tqdm(self.df_train["Region_ID"].unique(), leave=False)
        for idx, region in enumerate(pbar):
            if self.model_name in ["linear", "gam"]:
                self.create_feature_names(stages, dict_best_cei[region][0:3].tolist())
            elif self.ml_model:
                self.create_feature_names(stages, dict_selected_features[region])
            elif self.model_name in ["median"]:
                self.feature_names = [f"Median {self.target}"]
                self.last_year_yield_as_feature = False
                self.analogous_year_yield_as_feature = False
            elif self.model_name in ["analog"]:
                self.feature_names = [f"Analogous Year", f"Analogous Year Yield"]
                self.last_year_yield_as_feature = False
                self.median_yield_as_feature = False

            mask_train = self.df_train["Region_ID"] == region
            mask_test = self.df_test["Region_ID"] == region

            num_regions_in_cluster = self.df_train[mask_train]["Region"].unique()

            if self.cluster_strategy == "individual":
                region_name = self.df_train["Region"].unique()[idx]
                pbar.set_description(f"Fit/Predict for {region_name}")
                pbar.update()
            elif self.cluster_strategy in ["auto_detect", "single"]:
                pbar.set_description(f"Fit/Predict for group {idx + 1}")
                pbar.update()

            common_columns = (
                [self.target]
                + self.statistics_columns
                + self.feature_names
                + ["Region_ID"]
            )
            if self.check_yield_trend:
                common_columns += ["Detrended Model Type", "Detrended Model"]

            if self.last_year_yield_as_feature:
                common_columns += [f"Last Year {self.target}"]

            """ Train """
            # Filter dataframe based on region and self.feature_names
            df_region_train = self.df_train[mask_train]
            df_region_train = df_region_train[self.fixed_columns + common_columns]
            self.train(df_region_train, scaler)

            """ Predict """
            df_region_test = self.df_test[mask_test]
            df_region_test = df_region_test[self.fixed_columns + common_columns]
            experiment_id, df = self.predict(df_region_test, scaler)
            # df.reset_index(inplace=True)

            """ XAI """
            if self.do_xai:
                assert not self.estimate_ci, "Cannot perform XAI if estimate_ci is True"

                kwargs = {
                    "cluster_strategy": self.cluster_strategy,
                    "model": self.model,
                    "forecast_season": self.forecast_season,
                    "crop": self.crop,
                    "country": self.country,
                    "analysis_dir": self.dir_analysis,
                }
                df_xai = xai.explain(df_region_train, df_region_test, **kwargs)
                # df.set_index("Index", inplace=True)

            """ Output results to database """
            if not self.ml_model:
                model = self.model_name
            elif self.estimate_ci:
                model = self.model.estimator_
            else:
                model = self.model
            output.store(self.db_path, experiment_id, df, model, self.model_name)

    def create_ml_dataframe(self, df):
        """
        Create ML ready dataframe
        :param df:
        """
        _str = f"{self.country} {self.crop}"
        self.logger.info(f"Creating ML dataframe {_str}")

        # Convert from long to wide format
        df = df[
            ["Index", "Stage_ID", "CEI"]
            + self.fixed_columns
            + [self.target]
            + self.statistics_columns
        ]

        # For each combination of Index and Stage_ID, new columns have been
        # created and filled with the corresponding CEI value from the original dataframe
        # The new column names are a combination of the original Index and Stage_ID values
        # Use pivot_table to create the new columns and flatten the multi-index columns
        # HACK: Set missing values in self.target_col and self.statistics_columns to -1
        df.loc[:, [self.target] + self.statistics_columns] = df[
            [self.target] + self.statistics_columns
        ].fillna(-1)

        # HACK, replace later
        df.loc[:, "Area"] = df["Area"].fillna(-1)
        df = df.pivot_table(
            index=self.fixed_columns + [self.target] + self.statistics_columns,
            columns=["Index", "Stage_ID"],
            values="CEI",
        ).reset_index()

        # Reset self.target_col and self.statistics_columns to NaN
        df[[self.target] + self.statistics_columns] = df[
            [self.target] + self.statistics_columns
        ].replace(-1, np.nan)

        # Flatten the multi-index columns
        df.columns = [f"{i}_{j}" if j != "" else f"{i}" for i, j in df.columns]

        # HACK: Get feature name with GD4 in it to extract first and last stage id and name
        GD4_column = df[df.columns[df.columns.str.contains("GD4")]].columns
        # Select the longest string in GD4_column
        GD4_col = max(GD4_column, key=len)
        self.stage_info = stages.get_stage_information_dict(GD4_col, self.method)

        # Change column name
        # e.g. 'vDTR_7_6_5_4_3_2_1_37_36_35_34_33_32_31' to 'vDTR Mar 1-Oct 27'
        df = stages.update_feature_names(df, self.method)

        # Get all the columns apart from the fixed columns, target column and stats columns
        all_cei_columns = [
            col
            for col in df.columns
            if col not in self.fixed_columns + [self.target] + self.statistics_columns
        ]

        # Fill in any missing values with 0
        df.loc[:, all_cei_columns].fillna(0, inplace=True)

        if self.last_year_yield_as_feature:
            df = fe.compute_last_year_yield(df)

        if self.median_yield_as_feature:
            df = fe.compute_median_yield(
                df, self.all_seasons_with_yield, self.number_median_years
            )

        if self.lag_yield_as_feature:
            df = fe.compute_lag_yield(
                df, self.all_seasons_with_yield, self.number_lag_years
            )

        if self.analogous_year_yield_as_feature:
            df = fe.compute_analogous_yield(
                df, self.all_seasons_with_yield, self.number_median_years
            )

        # Create Region_ID column based on Region column category code
        df["Region"] = df["Region"].astype("category")
        if self.cluster_strategy == "single":
            df["Region_ID"] = 1
        elif self.cluster_strategy == "individual":
            df["Region_ID"] = df["Region"].cat.codes
        elif self.cluster_strategy == "auto_detect":
            clusters_assigned = fe.detect_clusters(df)
            # Merge the cluster labels with the original DataFrame
            df = df.merge(clusters_assigned, on="Region")

            # Region_ID should be type category
            df["Region_ID"] = df["Region_ID"].astype("category")

        return df

    def execute(self):
        """

        Returns:

        """
        """ Get dataframe including all simulation stages """
        # convert self.simulation_stages to a list of strings
        _stages = []
        for stage in self.simulation_stages:
            stage = stages.convert_stage_string(stage, to_array=False)
            _stages.append(stage)

        mask = self.df_results["Stage_ID"].isin(_stages)
        df = self.df_results[mask]

        """ Convert this dataframe into an ML ready format and save to disk """
        df = self.create_ml_dataframe(df)
        dir_output = (
            self.dir_analysis / self.country / self.crop / str(self.forecast_season)
        )
        os.makedirs(dir_output, exist_ok=True)
        df.to_csv(
            dir_output / f"{self.country}_{self.crop}_{self.forecast_season}.csv",
            index=False,
        )

        # cat_features should be converted to category type
        df[self.cat_features] = df[self.cat_features].astype("category")

        """  Heatmap of correlation of various features with yield at each time step"""
        # For each time step, determine the number of occurences of each feature
        # when it is most correlated with the target for a region
        # Use this to plot a heatmap showing which features are most correlated
        # with yield at each time step
        # Add a Country Region column to df and make it lower case
        df["Country Region"] = (
            df["Country"].astype(str) + " " + df["Region"].astype(str)
        ).str.lower()

        # Join with dg based on Country Region column, only keeping rows that are in df
        # Only use geometry column from self.dg
        self.dg_country = self.dg_country[["Country Region", "geometry"]].merge(
            df[["Country Region", self.correlation_plot_groupby]],
            on="Country Region",
            how="outer",
        )

        if self.correlation_plots:
            dict_kwargs = {}
            dict_kwargs["all_stages"] = self.all_stages
            dict_kwargs["target_col"] = self.target
            dict_kwargs["country"] = self.country
            dict_kwargs["crop"] = self.crop
            dict_kwargs["dir_output"] = (
                self.dir_analysis
                / self.country
                / self.crop
                / self.model_name
                / str(self.forecast_season)
            )
            dict_kwargs["forecast_season"] = self.forecast_season
            dict_kwargs["method"] = self.method
            dict_kwargs["national_correlation"] = self.national_correlation
            dict_kwargs["groupby"] = self.correlation_plot_groupby
            dict_kwargs["dg_country"] = self.dg_country
            dict_kwargs["combined_dict"] = self.combined_dict

            self.logger.info(f"Correlation plot for {self.country} {self.crop}")
            (
                dict_selected_features,
                dict_best_cei,
            ) = correlations.all_correlated_feature_by_time(df, **dict_kwargs)

        """ Separate into train and test datasets based on forecast_season """
        mask = df["Harvest Year"] == self.forecast_season
        self.df_train = df[~mask]
        self.df_test = df[mask]

        # Drop rows with missing values for self.target_col in df_train
        self.df_train = self.df_train.dropna(subset=[self.target])

        """ Groupby Region column and compute detrended yield """
        self.df_train[f"Detrended {self.target}"] = np.NaN
        self.df_train["Detrend Model"] = np.NaN
        self.df_train["Detrend Model Type"] = np.NaN
        if self.check_yield_trend:
            group_by = ["Region"]
            groups = self.df_train.groupby(group_by)
            for key, group in groups:
                if group[self.target].any():
                    detrended_data = trend.detrend_dataframe(
                        group, column_name=self.target
                    )
                    self.df_train.loc[
                        group.index, f"Detrended {self.target}"
                    ] = detrended_data.detrended_series
                    self.df_train.loc[
                        group.index, "Detrend Model"
                    ] = detrended_data.trend_model
                    self.df_train.loc[
                        group.index, "Detrend Model Type"
                    ] = detrended_data.model_type

        # 6. Exclude years without yields from df_train
        # self.df_train = self.df_train[
        #     self.df_train["Harvest Year"].isin(self.all_seasons_with_yield)
        # ]

        """ Run ML code for each stage """
        if self.run_ml:
            self.logger.info(f"Running ML for {self.country} {self.crop}")
            setup_stages = [self.simulation_stages]
            if self.run_latest_time_period:
                # setup_stages = stages.select_stages_for_ml(setup_stages, method="latest")
                setup_stages = [setup_stages[-1]]

            num_regions = len(self.df_train["Region_ID"].unique())
            pbar = tqdm(setup_stages)
            for stage in pbar:
                pbar.set_description(
                    f"ML {num_regions} regions, {len(setup_stages)} stages"
                )
                pbar.update()

                self.loop_ml(stage, dict_selected_features, dict_best_cei)

    def setup(self, forecast_season, model):
        """

        :param country:
        :param crop:
        :param forecast_season:
        :param model:
        """
        _str = f"{self.country} {self.crop} {model} {forecast_season}"
        self.logger.info(f"Setup {_str}")

        self.forecast_season = forecast_season
        self.model_name = model
        self.ml_model = self.parser.getboolean(self.model_name, "ML_model")
        self.model_names = ast.literal_eval(self.parser.get(self.country, "models"))
        self.optimize = self.parser.getboolean(self.country, "optimize")
        self.fraction_loocv = self.parser.getfloat(self.country, "fraction_loocv")
        self.all_seasons = self.df_results["Harvest Year"].unique()

        """ If not using a ML model then set XAI and CI to False """
        if not self.ml_model or self.model_name in ["linear", "gam", "merf"]:
            self.do_xai = False
            self.estimate_ci = False
            self.check_yield_trend = False
            self.estimate_ci_for_all = False
        else:
            self.do_xai = self.parser.getboolean("ML", "do_xai")
            self.estimate_ci = self.parser.getboolean("ML", "estimate_ci")
            self.estimate_ci_for_all = self.parser.getboolean(
                "ML", "estimate_ci_for_all"
            )
            self.check_yield_trend = self.parser.getboolean("ML", "check_yield_trend")

        # 1. Get all seasons with yield, these are the seasons for which Yield is available
        self.all_seasons_with_yield = self.df_results[
            self.df_results[self.target].notna()
        ]["Harvest Year"].unique()

        if self.method.endswith("_r"):
            if self.forecast_season == self.today_year:
                mask = self.df_results["Harvest Year"] == self.forecast_season
                self.all_stages = self.df_results[mask]["Stage_ID"].unique()
            else:
                self.all_stages = self.df_results["Stage_ID"].unique()
                # longest_stage = max(self.all_stages, key=len)
                # longest_stage = longest_stage.split("_")[::self.run_every_time_period]
                # # Select stages from all_stages which start with longest_stage
                # self.simulation_stages = [
                #     stage for stage in self.all_stages if stage.startswith(longest_stage)
                # ]
        else:
            raise NotImplementedError(f"Method {self.method} not implemented")

        # Create a list of numpy arrays, for each string in _stages,
        # create a numpy array by splitting by _ and converting to int
        # then store in self.simulation_stages
        # e.g. _stages = ['13_12_11', '13_12_11_10', '13_12_11_10_9']
        # then self.simulation_stages = [array([13, 12, 11]), array([13, 12, 11, 10]), array([13, 12, 11, 10, 9])]
        # Drop stages in self.all_stages that do not have _ in them
        self.all_stages = [element for element in self.all_stages if "_" in element]

        self.simulation_stages = [
            np.array([int(stage) for stage in s.split("_")]) for s in self.all_stages
        ]

        self.name_shapefile = self.parser.get(self.country, "boundary_file")
        self.admin_zone = self.parser.get(self.country, "admin_zone")
        self.dg = gp.read_file(
            self.dir_shapefiles / self.name_shapefile,
            engine="pyogrio",
        )

        if "ADMIN0" in self.dg.columns:
            # Hack rename Tanzania to United Republic of Tanzania
            self.dg["ADMIN0"] = self.dg["ADMIN0"].replace(
                "Tanzania", "United Republic of Tanzania"
            )

        # Rename ADMIN0 to ADM0_NAME and ADMIN1 to ADM1_NAME and ADMIN2 to ADM2_NAME
        self.dg = self.dg.rename(
            columns={
                "ADMIN0": "ADM0_NAME",
                "ADMIN1": "ADM1_NAME",
                "ADMIN2": "ADM2_NAME",
            }
        )

        if self.admin_zone == "admin_1":
            self.dg["Country Region"] = (
                self.dg["ADM0_NAME"] + " " + self.dg["ADM1_NAME"]
            )
        else:
            self.dg["Country Region"] = (
                self.dg["ADM0_NAME"] + " " + self.dg["ADM2_NAME"]
            )
        self.dg["Country Region"] = self.dg["Country Region"].str.lower()
        self.dg_country = self.dg[
            self.dg["ADM0_NAME"].str.lower().str.replace(" ", "_") == self.country
        ]

    def read_data(self, country, crop, season):
        """

        Args:
            country:
            crop:
            season:

        Returns:

        """
        self.logger.info(f"Reading data for {country} {crop} {season}")

        self.country = country
        self.crop = crop

        """Figure out path to file containing CEI data"""
        admin_zone = self.parser.get(country, "admin_zone")
        country_str = country.title().replace("_", " ")
        crop_str = crop.title().replace("_", " ")

        dir_statistics = self.dir_output / "cei" / "indices" / self.method / "global"
        os.makedirs(dir_statistics, exist_ok=True)
        file = (
            dir_statistics / f"{country_str}_{crop_str}_statistics_s1_{self.method}.csv"
        )

        """ Add area, yield and prod stats from either GEOGLAM or FEWSNET warehouse database"""
        if not os.path.exists(file) or self.update_input_file:
            _dir_country = (
                self.dir_output
                / "cei"
                / "indices"
                / self.method
                / admin_zone
                / country
                / crop
            )
            file_name = f"{country}_{crop}_s1*.csv"
            all_files = _dir_country.glob(file_name)
            # TODO ignore file with _2000 in its name
            all_files = [f for f in all_files if "_2000" not in f.name]

            self.df_results = pd.concat(
                (pd.read_csv(f) for f in all_files), ignore_index=True
            )

            self.df_results = stats.add_statistics(
                self.dir_condition / "yield",
                self.df_results,
                country_str,
                crop_str,
                admin_zone,
                self.statistics_columns + [self.target],
                self.method,
            )
            """ Add information on starting and ending time period for each stage"""
            self.df_results = stages.add_stage_information(self.df_results, self.method)

            self.df_results.to_csv(file, index=False)
        else:
            self.df_results = pd.read_csv(file)


def loop_execute(inputs):
    """

    Args:
        inputs:

    Returns:

    """
    from pycallgraph2 import Config, PyCallGraph, GlobbingFilter
    from pycallgraph2.output import GraphvizOutput

    graphviz = GraphvizOutput()
    graphviz.output_file = "geocif_visualization.png"
    plt.rcParams["figure.dpi"] = 600
    config = Config(max_depth=5)
    config.trace_filter = GlobbingFilter(
        exclude=[
            "pycallgraph.*",
        ]
    )

    with PyCallGraph(output=graphviz, config=config):
        country, crop, season, model, logger, parser = inputs

        obj = Geocif(logger=logger, parser=parser)
        obj.read_data(country, crop, season)

        # Setup metadata and run ML code
        obj.setup(season, model)
        if obj.simulation_stages:
            obj.execute()


def gather_inputs(parser):
    """

    Args:
        parser:

    Returns:

    """
    countries = ast.literal_eval(parser.get("DEFAULT", "countries"))

    """ Create a list of parameters over which to run the model"""
    all_inputs = []
    for country in countries:
        for crop in ast.literal_eval(parser.get(country, "crops")):
            for season in ast.literal_eval(parser.get(country, "forecast_seasons")):
                for model in ast.literal_eval(parser.get(country, "models")):
                    all_inputs.append([country, crop, season, model])

    return all_inputs


def execute_models(inputs, logger, parser):
    """
    Executes the model either in parallel or serially based on configuration.

    Args:
        inputs (list): The input data for model execution.
        logger (logging.Logger): Logger for tracking execution details
        parser (configparser.ConfigParser): Configuration file parser

    Returns:

    """
    do_parallel = parser.getboolean("DEFAULT", "do_parallel")

    # Add logger and parser to each element in inputs
    inputs = [item + [logger, parser] for item in inputs]

    if do_parallel:
        cpu_count = int(mp.cpu_count() * 0.3)
        with mp.Pool(cpu_count) as pool:
            pool.map(loop_execute, inputs)
    else:
        for inputs in tqdm(inputs, desc="Executing ML models"):
            loop_execute(inputs)

    logger.info("======================================")
    logger.info("\tCompleted all model executions")
    logger.info("======================================")


def main(logger, parser):
    """

    Args:
        logger:
        parser:

    Returns:

    """
    logger.info("===========================")
    logger.info("\tStarting GEOCIF")
    logger.info("===========================")

    inputs = gather_inputs(parser)
    execute_models(inputs, logger, parser)


def run(path_config_files=[Path("../config/geocif.txt")]):
    logger, parser = log.setup_logger_parser(path_config_files)
    main(logger, parser)


if __name__ == "__main__":
    run()



import multiprocessing as mp

import numpy as np
import optuna
from catboost import CatBoostRegressor
from sklearn.metrics import root_mean_squared_error
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def loocv(
    model,
    df,
    loocv_var,
    feature_names,
    target_col,
    fraction_loocv=1.0,
    cat_features=[],
    trial_id=0,
):
    """
    Perform Leave-One-Out Cross Validation (LOOCV)
    :param model: CatBoostRegressor, CatBoost model
    :param df: pd.DataFrame, training data
    :param loocv_var: str, variable to perform LOOCV on
    :param feature_names: list, list of feature names
    :param target_col: str, target column name
    :param fraction_loocv: float, fraction of unique values to perform LOOCV on
    :param cat_features: list, list of categorical feature names
    :return: float, average RMSE
    """
    rmse_values = []

    X = df[feature_names + cat_features]
    y = df[target_col]

    # Perform LOOCV based on precentage of loocv_var
    # Find unique values
    unique_values = df[loocv_var].unique()
    num_to_select = int(len(unique_values) * fraction_loocv)
    # Randomly select X% of the unique values without replacement
    selected_values = np.random.choice(unique_values, size=num_to_select, replace=False)
    pbar = tqdm(selected_values, leave=False)
    for idx, var in enumerate(pbar):
        pbar.set_description(f"Trial {trial_id}, LOOCV {var}")
        pbar.update()

        train_index = df[df[loocv_var] != var].index
        val_index = df[df[loocv_var] == var].index

        X_train, X_val = X.loc[train_index], X.loc[val_index]
        y_train, y_val = y.loc[train_index], y.loc[val_index]

        # Train the model
        model.fit(X_train, y_train)

        # Make predictions
        y_pred = model.predict(X_val)

        # Calculate RMSE
        rmse = root_mean_squared_error(y_val, y_pred)
        rmse_values.append(rmse)

    # Compute average MSE
    average_rmse = np.mean(rmse_values)

    return average_rmse


def optuna_objective(model, df, feature_names, target_col, cat_features=[]):
    """

    Args:
        model:
        df:
        feature_names:
        target_col:
        cat_features:

    Returns:

    """
    X = df[feature_names + cat_features]
    y = df[target_col]

    # Divide the data into training and validation sets
    train_X, val_X, train_y, val_y = train_test_split(X, y, test_size=0.2, random_state=0)

    model.fit(
        train_X,
        train_y,
        cat_features=cat_features,
        eval_set=(val_X, val_y),
        early_stopping_rounds=100,
        use_best_model=True,
        verbose=False,
    )

    # Make predictions
    val_preds = model.predict(val_X)

    # Evaluate predictions
    rmse = root_mean_squared_error(val_y, val_preds)

    return rmse


def optimized_model(
    model_name,
    df,
    use_loocv,
    loocv_var,
    feature_names,
    target_col,
    fraction_loocv,
    cat_features=[],
    seed=0,
):
    """
    Train CatBoost model using Optuna hyperparameter optimization
    :param model_name: str, 'CatBoost' or 'XGBoost'
    :param df: pd.DataFrame, training data
    :param loocv_var: str, 'Harvest Year'
    :param feature_names: list, list of feature names
    :param target_col: str, target column name
    :param fraction_loocv: float, fraction of unique values to perform LOOCV on
    :param cat_features: list, list of categorical feature names
    :param seed: int, random seed
    """
    # Define objecive function for optuna Hyperparameter tuning
    def _optuna_objective(trial):
        try:
            if model_name == "catboost":
                params = {
                    "depth": trial.suggest_int("depth", 1, 7),
                    "learning_rate": trial.suggest_float("learning_rate", 0.01, 0.1),
                    "iterations": trial.suggest_int("iterations", low=1000, high=5000, step=500),
                    "subsample": trial.suggest_float("subsample", 1.0, 1.0),
                    "random_strength": trial.suggest_float("random_strength", 0.3, 1.0),
                    "reg_lambda": trial.suggest_float("reg_lambda", 1e-8, 1.0),
                    "loss_function": "MAPE",
                    "early_stopping_rounds": 50,
                    "random_seed": seed,
                    "verbose": False,
                }

                # Fit the optuna model
                optuna_model = CatBoostRegressor(**params, cat_features=cat_features)
            else:
                raise NotImplementedError

            if use_loocv:
                trial_id = trial.number
                error_metric = loocv(
                    optuna_model,
                    df,
                    loocv_var,
                    feature_names,
                    target_col,
                    fraction_loocv,
                    cat_features,
                    trial_id,
                )
            else:
                error_metric = optuna_objective(
                    optuna_model, df, feature_names, target_col, cat_features
                )

            return error_metric
        except Exception as e:
            print(f"Trial failed with exception: {e}")
            return np.inf  # Assign a high cost to failed trials

    try:
        # Optimize hyperparameters
        n_trials = min(20, int(mp.cpu_count() * 0.9))
        optuna.logging.set_verbosity(optuna.logging.WARNING)  # Disable verbose
        sampler = optuna.samplers.TPESampler(seed=seed)
        study = optuna.create_study(sampler=sampler, direction="minimize")
        study.optimize(_optuna_objective, n_trials=n_trials, n_jobs=int(mp.cpu_count() * 0.4))
        if study.best_trial is None:
            raise ValueError("Optimization failed to complete any trials.")
        hyperparams = study.best_trial.params

    except Exception as e:
        print(f"Optimization failed: {e}")
        hyperparams = {
            "depth": 6,
            "learning_rate": 0.01,
            "iterations": 10,
            "subsample": 1.0,
            "random_strength": 0.5,
            "reg_lambda": 0.001,
            "loss_function": "MAPE",
            "early_stopping_rounds": 50,
            "random_seed": seed,
            "verbose": False,
        }

    # Model Initialization & Training
    if model_name == "catboost":
        model = CatBoostRegressor(**hyperparams, cat_features=cat_features)
    else:
        raise NotImplementedError

    return hyperparams, model


def auto_train(
    cluster_strategy,
    model_name,
    use_loocv,
    loocv_var,
    df_train,
    X_train,
    y_train,
    feature_names,
    target_col,
    optimize=False,
    fraction_loocv=1.0,
    cat_features=[],
    monotonic_fatures=[],
    seed=0,
):
    """
    Train CatBoost model using Optuna hyperparameter optimization

    :param cluster_strategy: str, 'individual' or 'auto_detect' or 'single'
    :param model_name: str, 'CatBoost' or 'XGBoost'
    :param stage: int, 'Dekad'
    :param loocv_var: str, 'Harvest Year'
    :param df_train: pd.DataFrame, training data
    :param feature_names: list, list of feature names
    :param target_col: str, target column name
    :param optimize: bool, whether to optimize hyperparameters
    :param fraction_loocv: float, fraction of unique values to perform LOOCV on
    :param cat_features: list, list of categorical feature names
    :param seed: int, random seed
    """
    if optimize:
        hyperparams, model = optimized_model(
            model_name,
            df_train,
            use_loocv,
            loocv_var,
            feature_names,
            target_col,
            fraction_loocv,
            cat_features,
            seed,
        )
    else:
        hyperparams = {}

        if model_name in ["catboost", "merf"]:
            hyperparams = {
                "depth": 6,
                "learning_rate": 0.01,
                "iterations": 5000,
                "subsample": 1.0,
                "random_strength": 0.5,
                "reg_lambda": 0.001,
                "loss_function": "MAPE",
                "early_stopping_rounds": 50,
                "random_seed": seed,
                "verbose": False,
            }
            if model_name == "catboost":
                model = CatBoostRegressor(**hyperparams, cat_features=cat_features)
            elif model_name == "merf":
                from merf import MERF

                hyperparams["iterations"] = 1000
                regr = CatBoostRegressor(**hyperparams, cat_features=cat_features)
                model = MERF(regr, max_iterations=10)
        elif model_name == "linear":
            from sklearn.linear_model import LassoCV

            model = LassoCV(cv=5, random_state=42)
        elif model_name == "gam":
            from pygam import LinearGAM

            model = LinearGAM(n_splines=25, spline_order=3).gridsearch(
                X_train.values, y_train.values, lam=np.logspace(-3, 3, 11)
            )
        elif model_name == "xgboost":
            raise NotImplementedError
        else:
            raise ValueError(f"Unknown model name: {model_name}")

    return hyperparams, model


def estimate_ci(model):
    """
    Estimate confidence intervals for the model
    using MapieRegressor
    Args:
        model:

    Returns:

    """
    from mapie.regression import MapieRegressor

    model = MapieRegressor(model, n_jobs=1)

    return model
